/*--------------------------------------------------------------------------------*
 *                                                                                *
 * CodecPro Incorporated                                                          *
 * 905-2160 Terry-Fox                                                             *
 * Laval (Quebec)                                                                 *
 * Canada, H7T 3B7                                                                *
 *                                                                                *
 * support@codecpro.com                                                           *
 * www.codecpro.com                                                               *
 *                                                                                *
 * This material is subject to continuous developments and improvements. All      *
 * warranties implied or expressed, including but not limited to, implied         *
 * warranties of merchantability, fitness for purpose, condition of title,        *
 * noninfringement, are excluded. In no event shall CodecPro Incorporated and its *
 * suppliers be liable for any special, indirect or consequential damages or any  *
 * damages whatsoever arising out of or in connection with the use of this        *
 * information. The foregoing disclaimer shall apply to the maximum extent        *
 * permitted by applicable law, even if any remedy fails its essential purpose.   *
 *--------------------------------------------------------------------------------*/

/*=================================================================================
   EVS Codec 3GPP TS26.442 (see disclaimer.h for version number)
 =================================================================================*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "mime_voip.h"

#ifndef _WIN32
//#include <netinet/in.h>
#include <stdint.h>
#else
#include <Winsock2.h>

typedef unsigned short     uint16_t;
typedef signed short       int16_t;
typedef unsigned int       uint32_t;
typedef signed int         int32_t;
typedef unsigned __int64   uint64_t;
typedef signed __int64     int64_t;
#endif

#ifdef _MSC_VER
#pragma warning( disable : 4996 )
#endif


#define EVS_VOIP_HEADER_SIZE_STORAGE          12
#define MAX_OCTETS_PER_FRAME                  2560/8

/*
 * Structures
 */

/* main handle */
struct __MIME
{
    FILE * file;
};

/*
 * Functions
 */

MIME_ERROR
MIME_Reader_Open(MIME_HANDLE* phMIME, FILE * filename)
{
   /* create handle */
    *phMIME = (MIME_HANDLE) calloc(1, sizeof(struct __MIME) );
    if ( !phMIME )
        return MIME_MEMORY_ERROR;
    memset(*phMIME, 0, sizeof(struct __MIME));

    /* associate file stream */
    (*phMIME)->file = filename;
    if( (*phMIME)->file == NULL )
    {
        return MIME_FILE_NOT_FOUND;
    }

    return MIME_NO_ERROR;
}

MIME_ERROR
MIME_ReadVoipFrame_compact(MIME_HANDLE const hMime,
                           unsigned char * const serial,
                           short * const num_octets,
                           unsigned short * const rtpSequenceNumber,
                           unsigned int * const rtpTimeStamp,
                           unsigned int * const rcvTime_ms)
{
    /*
        The format of the VoIP packet in the file is as such:
        Octets 0  to 1      : rtpPacketPayloadSize (MSB fist)
        Octets 2  to 5      : rtpPacketArrivalTime (MSB fist)
        Octets 6  to 7      : rtpSequenceNumber (MSB fist)
        Octets 8  to 11     : rtpTimeStamp (MSB fist)
        Octets 12 to (rtpPacketPayloadSize+12) : rtpPayload

        The format of the VoIP packet to EVS decoder is as such:
        (the packet to EVS decoder is packed in voip_client.c)
        Octets 0  to 3      : System Time (MSB fist)
        Octets 4  to 5      : rtpPacketPayloadSize (MSB fist)
        Octets 6  to 9      : rtpPacketArrivalTime (MSB fist)
        Octets 10 to 11     : rtpSequenceNumber (MSB fist)
        Octets 12 to 15     : rtpTimeStamp (MSB fist)
        Octets 16 to (rtpPacketPayloadSize+16) : rtpPayload

    */

    int16_t rtpPayloadSize;

    // read evs octet header from file
    if(fread(serial, sizeof(unsigned char), EVS_VOIP_HEADER_SIZE_STORAGE, hMime->file) != EVS_VOIP_HEADER_SIZE_STORAGE)
    {
        if(feof( hMime->file) != 0)
        {
            return MIME_EOF;
        }
        fprintf(stderr, "INPUT FILE too small\n");
        return MIME_READ_ERROR;
    }

    /* RTP payload size */
    rtpPayloadSize = (serial[0] << 8) + serial[1];

    *num_octets =  rtpPayloadSize;

    if(rtpPayloadSize < 0 || rtpPayloadSize > MAX_OCTETS_PER_FRAME)
        return MIME_READ_ERROR;

    /* RTP packet arrival time */
    *rcvTime_ms = (serial[2] << 24) + (serial[3] << 16) + (serial[4] << 8) + serial[5] ;

    /* RTP sequence number */
    *rtpSequenceNumber = (serial[6] << 8) + serial[7];

    /* RTP timestamp */
    *rtpTimeStamp = (serial[8] << 24) + (serial[9] << 16) + (serial[10] << 8) + serial[11];
    //
    // patch to convert timer ticks to ms
    #if 0
    *rtpTimeStamp >>= 4;    // in ms so divide timer tick (based on 16KHZ) by 16
    serial[8] = (*rtpTimeStamp >> 24) & 0xff;
    serial[9] = (*rtpTimeStamp >> 16) & 0xff;
    serial[10] = (*rtpTimeStamp >> 8) & 0xff;
    serial[11] = *rtpTimeStamp & 0xff;
    #endif
    //

    /* RTP payload size */
    if(rtpPayloadSize < 1)
    {
        fprintf(stderr, "RTP payload size too small: %u\n", rtpPayloadSize);
        //return MIME_READ_ERROR;
    }

    /* read RTP payload from file */
    if((int16_t)fread(serial, sizeof(unsigned char), rtpPayloadSize, hMime->file) != rtpPayloadSize)
    {
        fprintf(stderr, "Premature end of file\n");
        return MIME_READ_ERROR;
    }
    
    return MIME_NO_ERROR;
}

MIME_ERROR
MIME_Reader_Close(MIME_HANDLE* phMIME)
{
    if(phMIME == NULL || *phMIME == NULL)
        return MIME_NO_ERROR;

    free( *phMIME );
    *phMIME = NULL;
    phMIME = NULL;
    return MIME_NO_ERROR;
}

