/*--------------------------------------------------------------------------------*
 *                                                                                *
 * CodecPro Incorporated                                                          *
 * 905-2160 Terry-Fox                                                             *
 * Laval (Quebec)                                                                 *
 * Canada, H7T 3B7                                                                *
 *                                                                                *
 * support@codecpro.com                                                           *
 * www.codecpro.com                                                               *
 *                                                                                *
 * This material is subject to continuous developments and improvements. All      *
 * warranties implied or expressed, including but not limited to, implied         *
 * warranties of merchantability, fitness for purpose, condition of title,        *
 * noninfringement, are excluded. In no event shall CodecPro Incorporated and its *
 * suppliers be liable for any special, indirect or consequential damages or any  *
 * damages whatsoever arising out of or in connection with the use of this        *
 * information. The foregoing disclaimer shall apply to the maximum extent        *
 * permitted by applicable law, even if any remedy fails its essential purpose.   *
 *--------------------------------------------------------------------------------*/

/*=================================================================================
      EVS Codec 3GPP TS26.442 (see disclaimer.h for version number)
 =================================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "../evs_if.h"
#include "./rtpdump.h"

#include "mime_voip.h"
#include "../io_dec.h"

/*------------------------------------------------------------------------------------------*
 * Local constants
 *------------------------------------------------------------------------------------------*/

#define L_FRAME48k                            960       /* Frame size in samples at 48kHz */

#define VERBOSE 0


static const Word32 AMRWB_IO_mode2rate[16] = {
    6600,    // 0 - AMRWB_IO_6600
    8850,    // 1 - AMRWB_IO_8850
    12650,   // 2 - AMRWB_IO_1265
    14250,   // 3 - AMRWB_IO_1425
    15850,   // 4 - AMRWB_IO_1585
    18250,   // 5 - AMRWB_IO_1825
    19850,   // 6 - AMRWB_IO_1985
    23050,   // 7 - AMRWB_IO_2305
    23850,   // 8 - AMRWB_IO_2385
    1750,    // 9 - AMRWB_IO_SID
    -1,      // 10 - AMRWB_IO_FUT1
    -1,      // 11 - AMRWB_IO_FUT2
    -1,      // 12 - AMRWB_IO_FUT3
    -1,      // 13 - AMRWB_IO_FUT4
    0,       // 14 - SPEECH_BAD
    0        // 15 - NO_DATA
    };


static const Word32 PRIMARY_mode2rate[16] = {
    2800,   // 0 - PRIMARY_2800
    7200,   // 1 - PRIMARY_7200
    8000,   // 2 - PRIMARY_8000
    9600,   // 3 - PRIMARY_9600
    13200,  // 4 - PRIMARY_13200
    16400,  // 5 - PRIMARY_16400
    24400,  // 6 - PRIMARY_24400
    32000,  // 7 - PRIMARY_32000
    48000,  // 8 - PRIMARY_48000
    64000,  // 9 - PRIMARY_64000
    96000,  // 10 - PRIMARY_96000
    128000, // 11 - PRIMARY_128000
    2400,   // 12 - PRIMARY_SID
    -1,     // 13 - PRIMARY_FUT1
    0,      // 14 - SPEECH_LOST
    0       // 15  -NO_DATA
   };

Word32 cprtp_getFrameSize (UWord8 header)
{
   Word32 bitrate;

   if( header & 0x20 )
   {
      bitrate = AMRWB_IO_mode2rate[header & 0x0F];
   }
   else
   {
      bitrate = PRIMARY_mode2rate[ header & 0x0F ];
   }
   return ((bitrate/50) + 7)>>3;
}

/*------------------------------------------------------------------------------------------*
 * decodeVoip()
 *
 * Main function for EVS decoder with VOIP mode
 *------------------------------------------------------------------------------------------*/

Word16 decodeVoipMime(
    void *hDecState,
    void *hDecScratch,
    evs_decoderOption *decoderOption,
    FILE *f_stream,
    FILE *f_synth,
    Word16 sdp_hf_only,
    const Word16 quietMode,
    const Word16 noDelayCmp
)
{
    /* input/output */
    MIME_HANDLE     hMime = NULL;
    MIME_ERROR      mime_err = MIME_NO_ERROR;

    Word16          dec_delay, zero_pad;

    UWord8          au[(2560/8)+1+16];    // max frame size + frame header + header for RTP info
    Word16          auSize;
    UWord32         nextPacketRcvTime_ms = 0;
    UWord32         systemTime_ms = 0;
    UWord16         rtpSequenceNumber;
    UWord32         rtpTimeStamp;

    Word16          pcmBuf[L_FRAME48k] = {0};

    evs_streamInfo  streamInfo;
    Word32          nSamples;
    Word32          cpdec_err;
    Word16          bfi = 0;

    Word32          packet_cntr = 0;     // packet counter

    // the next 2 variables will hold data that needs to be transmitted to the remote
    // encoder for optimum error concealment recovery. See TS 26.114 clause 10.2, RTCP-APP
    // signaling adaptation requests.
    Word16          optimum_offset; // locally computed optimum Redundent Frame Offset
    Word16          FEC_hi;         // locally computed Frame Erasure Indicator

    // next set of variables are for rtpdump
    EVS_RTPDUMP_DEPACKER rtpdumpDepacker;
    EVS_RTPDUMP_DEPACKER_ERROR rtpdumpDepackerError = EVS_RTPDUMP_DEPACKER_NO_ERROR;
    UWord8*    auPtr = &au[17];
    uint8_t    toc;

    Word32 nFramesFeed = 0;



    rtpdumpDepacker.rtpdump = NULL;
    if( decoderOption->bitstreamformat == VOIP_RTPDUMP )
    {
       /* init rtpdump reader and open input file */
        rtpdumpDepackerError = EVS_RTPDUMP_DEPACKER_open(&rtpdumpDepacker, f_stream, sdp_hf_only != 0);
        if(rtpdumpDepackerError != EVS_RTPDUMP_DEPACKER_NO_ERROR)
        {
            fprintf(stderr,"error in EVS_RTPDUMP_DEPACKER_open(): %d\n", rtpdumpDepackerError);
            EVS_RTPDUMP_DEPACKER_close(&rtpdumpDepacker);
            return -1;
        }
    }
    else
    {
       /* init MIME reader and open input file */
       mime_err = MIME_Reader_Open(&hMime, f_stream);
       if(mime_err != MIME_NO_ERROR)
       {
           fprintf(stderr,"error in MIME_Reader_Open(): %d\n", mime_err);
           MIME_Reader_Close(&hMime);
           return -1;
       }
   }


    bfi = 0;
    decoderOption->jbmSafetyMargin = 60; /* allowed delay reserve in addition to network jitter to reduce late-loss [milliseconds] */
    au[16] = au[17] = au[18] = au[19] = 0; // not relevent for now, just set to 0
    D_IF_evs_getFrameInfo (hDecState, decoderOption, &au[16], &streamInfo);

    /* calculate the delay compensation to have the decoded signal aligned with the original input signal */
    /* the number of first output samples will be reduced by this amount */
    if (noDelayCmp == 0) {
        /* get the delay compensation to have the decoded signal aligned with the original input signal */
        /* the number of first output samples will be reduced by this amount */
        dec_delay = streamInfo.dec_delay;
    }
    else {
        dec_delay = 0;
    }
    zero_pad = dec_delay;

    if( quietMode == 0 )
    {
        fprintf( stdout, "\n------ Running the decoder ------\n\n" );
        fprintf( stdout, "Packets processed:       " );
    }
    else
    {
        fprintf( stdout, "\n-- Start the decoder (quiet mode) --\n\n" );
    }

    /* read the first packet */
    if( decoderOption->bitstreamformat == VOIP_RTPDUMP )
    {
        auPtr = &au[17]; /* might have been set to RTP packet in prev call */
        rtpdumpDepackerError = EVS_RTPDUMP_DEPACKER_readNextFrame(&rtpdumpDepacker, &rtpSequenceNumber, &rtpTimeStamp,
                                                                  &nextPacketRcvTime_ms,
                                                                  &toc,
                                                                  &auPtr, (uint16_t*)&auSize);
        /* EVS RTP payload format has timescale 16000, JBM uses 1000 internally */
        rtpTimeStamp = rtpTimeStamp / 16;
    }
    else
    {
        mime_err = MIME_ReadVoipFrame_compact(hMime, &au[16], &auSize,
                                         &rtpSequenceNumber, &rtpTimeStamp, &nextPacketRcvTime_ms);
    }
    if(mime_err != MIME_NO_ERROR)
    {
        fprintf(stderr,"failed to read first RTP packet\n");
        EVS_RTPDUMP_DEPACKER_close(&rtpdumpDepacker);
        MIME_Reader_Close(&hMime);
        return -1;
    }

    /* main receiving/decoding loop */
    for( ; ; )
    {
        /*
            The format of the VoIP packet to EVS decoder is as such:
            Octets 0  to 3      : System Time (MSB fist)
            Octets 4  to 5      : rtpPacketPayloadSize (MSB fist)
            Octets 6  to 9      : rtpPacketArrivalTime (MSB fist)
            Octets 10 to 11     : rtpSequenceNumber (MSB fist)
            Octets 12 to 15     : rtpTimeStamp (MSB fist)
            Octets 16 to (rtpPacketPayloadSize+16) : rtpPayload
        */
        // update System Time to packet 
        au[0] = (systemTime_ms >> 24) & 0xff;
        au[1] = (systemTime_ms >> 16) & 0xff;
        au[2] = (systemTime_ms >> 8) & 0xff;
        au[3] = systemTime_ms & 0xff;
        // add rtp packet arrival time 
        au[6] = (nextPacketRcvTime_ms >> 24) & 0xff;
        au[7] = (nextPacketRcvTime_ms >> 16) & 0xff;
        au[8] = (nextPacketRcvTime_ms >> 8) & 0xff;
        au[9] = nextPacketRcvTime_ms & 0xff;

        /* read all packets with a receive time smaller than the system time */
        if ( nextPacketRcvTime_ms <= systemTime_ms )
        {
            //
            // feed packet to codecLib and read next packet
            //

            packet_cntr++;

            // feed previously read packet
            // add rtp packet payload size 
            au[4] = (auSize >> 8) & 0xff;
            au[5] = auSize & 0xff;
            // add rtp packet sequence number
            au[10] = (rtpSequenceNumber >> 8) & 0xff;
            au[11] = rtpSequenceNumber & 0xff;
            // add rtp packet time stamp
            au[12] = (rtpTimeStamp >> 24) & 0xff;
            au[13] = (rtpTimeStamp >> 16) & 0xff;
            au[14] = (rtpTimeStamp >> 8) & 0xff;
            au[15] = rtpTimeStamp & 0xff;

            if( decoderOption->bitstreamformat == VOIP_RTPDUMP )
            {
                Word32 frameSize;

                // build frame header
                /*0 1 2 3 4 5 6 7   MS-bit ---> LS-bit
                +-+-+-+-+-+-+-+-+
                |H|F|E|x| brate |
                +-+-+-+-+-+-+-+-+
                  where :
                     "E|x|  brate "  is the 6 bit "FT" -field
                     x is unused    if E=0, (should be 0 )
                     x is the Q-bit if E=1, Q==1(good), Q==0(bad, maybe bit errors in payload )
                     H,F  always   0 in RTP format.
                */
                frameSize  = cprtp_getFrameSize (toc);
                au[16] = toc;
                memcpy(&au[17], auPtr, frameSize);
            }

            //
            // when nextPacketRcvTime_ms <= systemTime_ms the EVS decoder feeds the RTP packet to the
            // EVS jitter buffer and will NOT output any PCM data.
            //

            cpdec_err = D_IF_evs_decode (
                        hDecState,
                        decoderOption,
                        au, 
                        pcmBuf, 
                        &nSamples,
                        bfi);

            if (cpdec_err != EVS_DEC_ERR_NONE)
            {
                printf("\nerror in D_IF_evs_decode_Voip: %8x", cpdec_err);
                EVS_RTPDUMP_DEPACKER_close(&rtpdumpDepacker);
                MIME_Reader_Close(&hMime);
                return -1;
            }

            ++nFramesFeed;

            /* read the next packet */
            if( decoderOption->bitstreamformat == VOIP_RTPDUMP )
            {
                auPtr = &au[17]; /* might have been set to RTP packet in prev call */
                rtpdumpDepackerError = EVS_RTPDUMP_DEPACKER_readNextFrame(&rtpdumpDepacker, &rtpSequenceNumber, &rtpTimeStamp,
                                                                          &nextPacketRcvTime_ms,
                                                                          &toc,
                                                                          &auPtr, (uint16_t*)&auSize);
                /* EVS RTP payload format has timescale 16000, JBM uses 1000 internally */
                rtpTimeStamp = rtpTimeStamp / 16;
            }
            else
            {
                mime_err = MIME_ReadVoipFrame_compact(hMime, &au[16], &auSize,
                                                    &rtpSequenceNumber, &rtpTimeStamp, &nextPacketRcvTime_ms);
            }

            if ((mime_err == MIME_EOF)
             || (rtpdumpDepackerError == EVS_RTPDUMP_DEPACKER_EOF))
            {
                /* finished reading */
                nextPacketRcvTime_ms = (unsigned int) -1;
            }
            else if ((mime_err != MIME_NO_ERROR)
                 || (rtpdumpDepackerError != EVS_RTPDUMP_DEPACKER_NO_ERROR))
            {
                fprintf(stderr,"failed to read RTP packet\n");
                EVS_RTPDUMP_DEPACKER_close(&rtpdumpDepacker);
                MIME_Reader_Close(&hMime);
                return -1;
            }
        }

        else
        {
            //
            // call the decoder with no payload every 20 ms to get PCM data
            //

            // rtp packet payload size, rtp packet sequence number and rtp packet time stamp
            // are not used by the decoder when nextPacketRcvTime_ms > systemTime_ms

            //
            // when the nextPacketRcvTime_ms > systemTime_ms the decoder will fetch encoded frames
            // from the jitter buffer, feed it to the decoder and output 20 ms of data.
            //
            cpdec_err = D_IF_evs_decode (
                        hDecState,
                        decoderOption,
                        au, 
                        pcmBuf, 
                        &nSamples,
                        bfi);

            if (cpdec_err != EVS_DEC_ERR_NONE)
            {
                printf("\nerror in D_IF_evs_decode_Voip: %8x", cpdec_err);
                EVS_RTPDUMP_DEPACKER_close(&rtpdumpDepacker);
                MIME_Reader_Close(&hMime);
                return -1;
            }

            D_IF_evs_getFrameInfo (hDecState, decoderOption, &au[16], &streamInfo); // get streamInfo.jitterBufferEmpty

            // these need to be transmitted using "RTCP-APP signaling adaptation requests" to the remote
            // encoder for optimum quality of two way conversation.
            optimum_offset = streamInfo.optimum_offset;
            FEC_hi = streamInfo.FEC_hi;

            if ( dec_delay == 0 )
            {
                fwrite( pcmBuf, sizeof(Word16), nSamples, f_synth );
            }
            else
            {
                if ( dec_delay <= (short)nSamples )
                {
                    fwrite( pcmBuf + dec_delay, sizeof(Word16), nSamples - dec_delay, f_synth );
                    dec_delay = 0;
                }
                else
                {
                    dec_delay -= nSamples;
                }
            }

            if( quietMode == 0 )
            {
                fprintf( stdout, "%-8d\b\b\b\b\b\b\b\b", packet_cntr);
            }

            /* we are finished when all packets have been received and jitter buffer is empty */
            /* also stop when the input file contains less than two frames, because JBM cannot calculate a delay value and won't start decoding */
            if( nextPacketRcvTime_ms == (UWord32)(-1))
            {
                if (streamInfo.jitterBufferEmpty || (nFramesFeed < 2) )
                    break;  // break out the for loop
            }

            systemTime_ms += 20;    // systemTime update takes effect only on the next loop
        }
    }

    /* add zeros at the end to have equal length of synthesized signals */
    memset(pcmBuf, 0, zero_pad*2 );

    fwrite( pcmBuf, sizeof(Word16), zero_pad, f_synth );

    if (quietMode == 0)
    {
        fprintf( stdout, "\n\n" );
        fprintf( stdout, "Decoding finished\n\n" );
    }
    else
    {
        fprintf( stdout, "Decoding of %d frames finished\n\n", packet_cntr);
    }
    fprintf( stdout, "\n\n" );

    /* free memory etc. */
    EVS_RTPDUMP_DEPACKER_close(&rtpdumpDepacker);
    MIME_Reader_Close(&hMime);
    return 0;
}

