/*--------------------------------------------------------------------------------*
 *                                                                                *
 * CodecPro Incorporated                                                          *
 * 905-2160 Terry-Fox                                                             *
 * Laval (Quebec)                                                                 *
 * Canada, H7T 3B7                                                                *
 *                                                                                *
 * support@codecpro.com                                                           *
 * www.codecpro.com                                                               *
 *                                                                                *
 * This material is subject to continuous developments and improvements. All      *
 * warranties implied or expressed, including but not limited to, implied         *
 * warranties of merchantability, fitness for purpose, condition of title,        *
 * noninfringement, are excluded. In no event shall CodecPro Incorporated and its *
 * suppliers be liable for any special, indirect or consequential damages or any  *
 * damages whatsoever arising out of or in connection with the use of this        *
 * information. The foregoing disclaimer shall apply to the maximum extent        *
 * permitted by applicable law, even if any remedy fails its essential purpose.   *
 *--------------------------------------------------------------------------------*/

/*=================================================================================
    EVS Codec 3GPP TS26.442 (see disclaimer.h for version number)
 =================================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include "evs_if.h"
#include "io_dec.h"
#include "wavefiletools.h"

#ifdef __GNUC__
#define BILLION 1000000000L
#define GET_TIME(time) clock_gettime(CLOCK_MONOTONIC, &time); 
#endif

/*------------------------------------------------------------------------------------------*
 * Main decoder function
 *------------------------------------------------------------------------------------------*/


Word32 main( Word32 argc, char *argv[] )
{
//#define STORE_INPUT_DATA
#ifdef STORE_INPUT_DATA
    FILE* storeInDataFile;
#endif

    FILE *f_stream;                     /* input bitstream file */
    FILE *f_synth;                      /* output synthesis file */
    Word16 dec_delay, zero_pad;
    Word16 quietMode = 0;
    Word16 noDelayCmp = 0;
    Word16 output[L_FRAME48k];           /* buffer for output synthesis */
    UWord8         bitstream[((MAX_BITS_PER_FRAME + 7) >> 3)+1];  // max MIME frame size
    // state variables
    void            *hDecState = 0;         // decoder state memory
    Word32          iDecSizeState = 0;      // size of decoder state memory in bytes
    void            *hDecScratch = 0;       // decoder scratch memory
    Word32          iDecSizeScratch = 0;    // size of decoder scratch memory in bytes
    // command line options
    // codec configuration and codec info structures
    evs_decoderOption decoderOption;
    evs_streamInfo  streamInfo;
    // work variables
    Word32          bitrate, i;
    Word32          outputSize;
    UWord16         headerSize;
    Word16          bfi = 0;
    Word32          frame_cnt = 0;              /* Counter of frames */
    Word16          sdp_hf_only = 0;

    Word32 ret_val = 0;
#ifdef __GNUC__
   Word32 cpu_duration_s, total_cpu_duration_s=0;
   Word32 cpu_duration_ns, total_cpu_duration_ns=0;
   struct timespec start;
   struct timespec finish;
#else
   clock_t start,finish;
   double cpu_duration, total_cpu_duration=0;
#endif


    Word32 hf_only = 0;  // set to 0 for both Compact and Header-Full format allowed, set to 1 for Header-Full only.

    for (i = 0; i < (((MAX_BITS_PER_FRAME + 7) >> 3)+1); i++)
    {
        bitstream[i] = 0;
    }

    /*------------------------------------------------------------------------------------------*
     * Allocate memory for static variables
     * Processing of command-line parameters
     * Decoder initialization
     *------------------------------------------------------------------------------------------*/

    // allocate decoder memory block and clear it to 0
    D_IF_evs_queryBlockSize(&iDecSizeState, &iDecSizeScratch);

    hDecState = calloc(iDecSizeState, 1);   // allocate and clear state memory
    if (!hDecState) {
        fprintf(stderr, "state memory allocation for decoder - malloc failed!\n");
        return(-1);
    }
    hDecScratch = calloc(iDecSizeScratch, 1);   // allocate and clear state memory
    if (!hDecScratch) {
        if (hDecState) free(hDecState);
        fprintf(stderr, "scratch memory allocation for decoder - malloc failed!\n");
        return(-1);
    }

    if (-1 == io_ini_dec( argc, argv, &f_stream, &f_synth,
                &quietMode, &noDelayCmp,
                &decoderOption.Opt_VOIP,
                &decoderOption.output_Fs,
                &decoderOption.bitstreamformat,
                &decoderOption.amrwb_rfc4867_flag,
                &sdp_hf_only
              ))
    {
        if (hDecState) free(hDecState);
        if (hDecScratch) free(hDecScratch);
        if (f_stream) fclose(f_stream);
        if (f_synth) fclose(f_synth);
        return (-1);
    }

    // Initialize decoder
    decoderOption.jbmSafetyMargin = 60;     // for VoIP option - not handled by io_ini_dec()
    if (EVS_DEC_ERR_NONE!= D_IF_evs_init(hDecState, hDecScratch, &decoderOption))
    {
        fprintf( stderr, "\n-- EVS decoder init failed --\n\n" );
        // make sure to call close to free malloc'ed memory
        D_IF_evs_close(hDecState);
        if (hDecState) free(hDecState);
        if (hDecScratch) free(hDecScratch);
        if (f_stream) fclose(f_stream);
        if (f_synth) fclose(f_synth);
        return (-1);
    }
    

    if( decoderOption.Opt_VOIP )
    {
        ret_val = decodeVoipMime(hDecState, hDecScratch, &decoderOption, f_stream, f_synth,
                    sdp_hf_only,
            quietMode,
            noDelayCmp
            );

        // make sure to call close to free malloc'ed memory
        D_IF_evs_close(hDecState);
        if (hDecState) free(hDecState);
        if (hDecScratch) free(hDecScratch);
        if (f_stream) fclose(f_stream);
        if (f_synth) CheckIfWave_Fclose(f_synth);
#ifdef STORE_INPUT_DATA
        if (storeInDataFile) fclose(storeInDataFile);
#endif

        if( quietMode == 0 )
            fprintf(stdout, "\n...Done... \n");

        if (ret_val != 0) return -1;
        else return 0;
    }


    /*------------------------------------------------------------------------------------------*
     * Regular EVS decoder with MIME bitstream
     *------------------------------------------------------------------------------------------*/

    headerSize = 1;

    /*------------------------------------------------------------------------------------------*
    * Allocate memory for static variables
    * Decoder initialization
    *------------------------------------------------------------------------------------------*/

    bitrate = streamInfo.bitrate = 0;     // init to 0
    bitstream[0] = bitstream[1] = bitstream[2] = bitstream[3] = 0; // not relevent for now, just set to 0
    D_IF_evs_getFrameInfo (hDecState, &decoderOption, bitstream, &streamInfo); // get delay

    if( noDelayCmp == 0 ){
        /* get the delay compensation to have the decoded signal aligned with the original input signal */
        /* the number of first output samples will be reduced by this amount */
        dec_delay = streamInfo.dec_delay;
    }
    else{
        dec_delay = 0;
    }
    zero_pad = dec_delay;

    /*------------------------------------------------------------------------------------------*
        * Loop for every packet  (frame) of bitstream data
        * - Read the bitstream packet
        * - Run the decoder
        * - Write the synthesized signal into output file
        *------------------------------------------------------------------------------------------*/

    if( quietMode == 0 ){
        fprintf( stdout, "\n------ Running the decoder ------\n\n" );
    }

#ifdef STORE_INPUT_DATA
    if ((storeInDataFile = fopen("storeInDataFile.evs", "wb")) == NULL)
    {
        fprintf(stderr, "Error: input bitstream file %s cannot be opened\n\n", "storeInDataFile.evs");
        return -1;
    }
    char buf[4];
    fwrite(EVS_MAGIC_NUMBER, sizeof(char), strlen(EVS_MAGIC_NUMBER), storeInDataFile);
    buf[0] = buf[1] = buf[2] = 0;
    buf[3] = 1;
    fwrite(&buf, sizeof(char), 4, storeInDataFile);
#endif

#ifdef __GNUC__
    total_cpu_duration_ns = 0;
    total_cpu_duration_s = 0;
#else
    start = clock();
#endif

    while (headerSize == fread(bitstream, sizeof(UWord8), headerSize, f_stream))
    {
        if (EVS_DEC_ERR_NONE != D_IF_evs_getFrameInfo(hDecState, &decoderOption, bitstream, &streamInfo)){
            fprintf(stderr, "EVS decoder - GetFrameInfo error!\n");
            // free resources
            // make sure to call close to free malloc'ed memory
            D_IF_evs_close(hDecState);
            if (hDecState) free(hDecState);
            if (hDecScratch) free(hDecScratch);
            if (f_stream) fclose(f_stream);
            if (f_synth) CheckIfWave_Fclose(f_synth);
            return (-1);
        }
        if ((streamInfo.frame_Size-headerSize) != (Word32)fread( &bitstream[headerSize], sizeof(UWord8), (streamInfo.frame_Size-headerSize), f_stream )){
            fprintf(stderr, "EVS decoder - last frame of file incomplete!\n");
            // free resources
            // make sure to call close to free malloc'ed memory
            D_IF_evs_close(hDecState);
            if (hDecState) free(hDecState);
            if (hDecScratch) free(hDecScratch);
            if (f_stream) fclose(f_stream);
            if (f_synth) CheckIfWave_Fclose(f_synth);
            return (-1);
        }

#ifdef STORE_INPUT_DATA
        fwrite(bitstream, sizeof(Word8), streamInfo.frame_Size, storeInDataFile);
#endif

#ifdef __GNUC__
        GET_TIME(start);
#endif

        if (EVS_DEC_ERR_NONE != D_IF_evs_decode(
                                hDecState,
                                &decoderOption,
                                &bitstream[0], 
                                output, 
                                &outputSize,
                                bfi)){
            fprintf(stderr, "EVS decoder error!\n");
            // free resources
            // make sure to call close to free malloc'ed memory
            D_IF_evs_close(hDecState);
            if (hDecState) free(hDecState);
            if (hDecScratch) free(hDecScratch);
            if (f_stream) fclose(f_stream);
            if (f_synth) CheckIfWave_Fclose(f_synth);

#ifdef STORE_INPUT_DATA
            if (storeInDataFile) fclose(storeInDataFile);
#endif
            return (-1);
        }

#ifdef __GNUC__
        GET_TIME(finish);
        cpu_duration_ns = finish.tv_nsec - start.tv_nsec;
        if (cpu_duration_ns < 0){
            cpu_duration_ns = cpu_duration_ns + BILLION;
        }
        total_cpu_duration_ns += cpu_duration_ns;

        if (total_cpu_duration_ns >= BILLION) {
            total_cpu_duration_s += 1;
            total_cpu_duration_ns -= BILLION;
        }

        cpu_duration_s = finish.tv_sec - start.tv_sec;
        if (cpu_duration_s > 1) printf( "\n STRANGE cpu_duration_s= %d.", (Word32)cpu_duration_s);
#endif
        /* write the synthesized signal into output file */
        /* do final delay compensation */
        if ( dec_delay == 0 ){
            fwrite( output, sizeof(short), outputSize, f_synth );
        }
        else{
            if ( dec_delay <= outputSize ){
                fwrite( &output[dec_delay], sizeof(short), outputSize - dec_delay, f_synth );
                dec_delay = 0;
            }
            else{
                dec_delay -= outputSize;
            }
        }
	    fprintf(stdout, "frame : %d \r", frame_cnt);
        frame_cnt++;
    }
    // end of decoding loop

    if (quietMode == 0){
        fprintf(stdout, "Decoding of %d frames finished\n\n", frame_cnt);
        fprintf( stdout, "\n\n" );
    }

    /* add zeros at the end to have equal length of synthesized signals */
    for (i=0 ; i<zero_pad ; i++) {
        output[i] = 0;
    }
    fwrite( output, sizeof(short), zero_pad, f_synth );


#ifdef __GNUC__

    if( quietMode == 0 )
    {
        printf("\n %s Decoder CPU load Time: %d.", argv[3], (Word32)total_cpu_duration_s);
        if (total_cpu_duration_ns < 100000000L)   printf("0");
        if (total_cpu_duration_ns < 10000000L)   printf("0");
        if (total_cpu_duration_ns < 1000000L)   printf("0");
        if (total_cpu_duration_ns < 100000L)   printf("0");
        if (total_cpu_duration_ns < 10000L)   printf("0");
        if (total_cpu_duration_ns < 1000L)   printf("0");
        if (total_cpu_duration_ns < 100L)   printf("0");
        if (total_cpu_duration_ns < 10L)   printf("0");
        printf( "%d", (Word32)total_cpu_duration_ns);
        printf( " s.\n");
    }




#else
    finish = clock();
    cpu_duration = (double) (finish - start);
    total_cpu_duration += cpu_duration;
    if (quietMode == 0)
        fprintf(stdout, "Decoding process CPU load Time: %4.4f sec ", (float)total_cpu_duration / CLOCKS_PER_SEC);
#endif

// free resources
    // make sure to call close to free malloc'ed memory
    D_IF_evs_close(hDecState);
    if (hDecState) free(hDecState);
    if (hDecScratch) free(hDecScratch);
    if (f_stream) fclose(f_stream);
    if (f_synth) CheckIfWave_Fclose(f_synth);
#ifdef STORE_INPUT_DATA
    if (storeInDataFile) fclose(storeInDataFile);
#endif

    if (quietMode == 0)
        fprintf(stdout, "\n...Done... \n");

    return 0;
}
