/*--------------------------------------------------------------------------------*
 *                                                                                *
 * CodecPro Incorporated                                                          *
 * 905-2160 Terry-Fox                                                             *
 * Laval (Quebec)                                                                 *
 * Canada, H7T 3B7                                                                *
 *                                                                                *
 * support@codecpro.com                                                           *
 * www.codecpro.com                                                               *
 *                                                                                *
 * This material is subject to continuous developments and improvements. All      *
 * warranties implied or expressed, including but not limited to, implied         *
 * warranties of merchantability, fitness for purpose, condition of title,        *
 * noninfringement, are excluded. In no event shall CodecPro Incorporated and its *
 * suppliers be liable for any special, indirect or consequential damages or any  *
 * damages whatsoever arising out of or in connection with the use of this        *
 * information. The foregoing disclaimer shall apply to the maximum extent        *
 * permitted by applicable law, even if any remedy fails its essential purpose.   *
 *--------------------------------------------------------------------------------*/

/*=================================================================================
      EVS Codec 3GPP TS26.442 (see disclaimer.h for version number)
 =================================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <wchar.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <sys/stat.h>
#include "evs_if.h"
#include "io_enc.h"

#ifdef __GNUC__
#define BILLION 1000000000L
#define GET_TIME(time) clock_gettime(CLOCK_MONOTONIC, &time); 
#endif
 
/*------------------------------------------------------------------------------------------*
 * Main encoder function
 *------------------------------------------------------------------------------------------*/


Word32 main( Word32 argc, char *argv[] )
{

    FILE *f_input;                                        /* MODE1 - input signal file */
    FILE *f_stream;                                       /* MODE1 - output bitstream file */
    Word16 input_frame, n_samples, encodedFrameSize;
    Word32 enc_delay;
    Word16 quietMode = 0;
    Word16 noDelayCmp = 0;
    Word16 data[L_FRAME48k];                               /* 'short' buffer for input signal */
    UWord8 pFrame[((MAX_BITS_PER_FRAME + 7) >> 3)+1];  // max MIME frame size

    // state variables
    void            *hEncState;                         // encoder state memory
    void            *hEncScratch;                       // encoder scratch memory
    Word32          iEncSizeState = 0;                  // size of encoder state memory in bytes
    Word32          iEncSizeScratch = 0;                // size of encoder scratch memory in bytes
    // codec configuration and codec info structures
    evs_encoderOption encoderOption;
    // work variables
    Word32          output_size;
    Word32          i;
    Word32          frame = 0;

#ifdef __GNUC__
    Word32 cpu_duration_s, total_cpu_duration_s=0;
    Word32   cpu_duration_ns, total_cpu_duration_ns=0;
   struct timespec start, finish;
#else
   clock_t start,finish;
   double cpu_duration, total_cpu_duration=0;
#endif


    /*------------------------------------------------------------------------------------------*
     * Allocate memory for static variables
     * Processing of command-line parameters
     * Encoder initialization
     *------------------------------------------------------------------------------------------*/

    // allocate encoder memory block
    E_IF_evs_queryBlockSize(&iEncSizeState, &iEncSizeScratch);

    hEncState = malloc(iEncSizeState);
    if (!hEncState)
    {
        fprintf(stderr, "state memory allocation for encoder - malloc failed!\n");
        return(-1);
    }
    hEncScratch = malloc(iEncSizeScratch);
    if (!hEncScratch)
    {
        fprintf(stderr, "scratch memory allocation for encoder - malloc failed!\n");
        return(-1);
    }

    if (-1 == io_ini_enc( argc, argv, &f_input, &f_stream, &quietMode, &noDelayCmp,
                &encoderOption.input_Fs,
                &encoderOption.total_brate,
                &encoderOption.bitstreamformat,
                &encoderOption.Opt_DTX_ON,
                &encoderOption.Opt_RF_ON,
                &encoderOption.rf_fec_offset,
                &encoderOption.rf_fec_indicator,
                &encoderOption.max_bwidth,
                &encoderOption.interval_SID
              ))
              return (-1);

    // Initialize encoder
    if (EVS_ENC_ERR_NONE!= E_IF_evs_init(hEncState, hEncScratch, &encoderOption))
    {
        fprintf( stderr, "\n-- EVS encoder init failed --\n\n" );
        E_IF_evs_close(hEncState);
        // make sure to call close to free malloc'ed memory
        if (hEncState) free(hEncState);
        if (hEncScratch) free(hEncScratch);
        if (f_input) fclose(f_input);
        if (f_stream) fclose(f_stream);
        return (-1);
    }
    
    input_frame = (short)(encoderOption.input_Fs / 50);

    /*------------------------------------------------------------------------------------------*
     * Compensate for encoder delay (bitstream aligned with input signal)
     * Compensate for the rest of codec delay (local synthesis aligned with decoded signal and original signal)
     *------------------------------------------------------------------------------------------*/
    E_IF_evs_getInfo(&encoderOption, &enc_delay, &encodedFrameSize);

    if ( noDelayCmp == 0 )
    {
        /* read samples and throw them away */
        if( fread(data, sizeof(short), enc_delay, f_input) != enc_delay )
        {
            fprintf(stderr, "Can not read the data from input file\n");
            return (-1);
        }
        /* the number of first output samples will be reduced by this amount */
    }


    /*------------------------------------------------------------------------------------------*
     * Loop for every frame of input data
     * - Read the input data
     * - Process switching files
     * - Run the encoder
     * - Write the parameters into output bitstream file
     *------------------------------------------------------------------------------------------*/

    if( quietMode == 0 ) {
        fprintf( stdout, "\n------ Running the encoder ------\n\n" );
    }
    //fprintf( stdout, "\n _MSC_VER = %d\n\n", _MSC_VER );

#ifdef __GNUC__
   total_cpu_duration_ns = 0;
   total_cpu_duration_s = 0;
#else
   start = clock();
#endif
    
    while( (n_samples = (short)fread(data, sizeof(short), input_frame, f_input)) > 0 ) {

        // if last frame of file is not even to frame length fill with 0
        for( i = n_samples; i < input_frame; i++ )        {
            data[i] = 0;
        }


#ifdef __GNUC__
        GET_TIME(start);
#endif

        if (EVS_ENC_ERR_NONE != E_IF_evs_encode(hEncState,
                        &encoderOption,
                        data, 
                        pFrame, 
                        &output_size)) {

            fprintf( stderr, " EVS - Encoder error\n");
        }
#ifdef __GNUC__
        GET_TIME(finish);
        cpu_duration_ns = finish.tv_nsec - start.tv_nsec;
        if (cpu_duration_ns < 0){
           cpu_duration_ns = cpu_duration_ns + BILLION;
        }
        total_cpu_duration_ns += cpu_duration_ns;

        if (total_cpu_duration_ns >= BILLION) {
           total_cpu_duration_s += 1;
           total_cpu_duration_ns -= BILLION;
        }

        cpu_duration_s = finish.tv_sec - start.tv_sec;
        if (cpu_duration_s > 1) printf( "\n STRANGE cpu_duration_s= %d.", (Word32)cpu_duration_s);
#endif

        fwrite (pFrame, sizeof(UWord8), output_size, f_stream);
        fflush( stdout );
        fflush( stderr );

        frame++;
            
    }
    if( quietMode == 0 ) {
        fprintf( stdout, "Encoding of %d frames finished\n\n", frame );
        fprintf( stdout, "Encoding finished\n\n" );
        fprintf( stdout, "\n\n" );
    }

#ifndef __GNUC__
    finish = clock();
    cpu_duration = (double) (finish - start);
    total_cpu_duration += cpu_duration;
#endif


#ifdef __GNUC__
    if( quietMode == 0 )
    {
        printf( "\n Encoder CPU load Time: %d.", (Word32)total_cpu_duration_s);
        if (total_cpu_duration_ns < 100000000L)   printf( "0");
        if (total_cpu_duration_ns < 10000000L)   printf( "0");
        if (total_cpu_duration_ns < 1000000L)   printf( "0");
        if (total_cpu_duration_ns < 100000L)   printf( "0");
        if (total_cpu_duration_ns < 10000L)   printf( "0");
        if (total_cpu_duration_ns < 1000L)   printf( "0");
        if (total_cpu_duration_ns < 100L)   printf( "0");
        if (total_cpu_duration_ns < 10L)   printf( "0");
        printf( "%d", (Word32)total_cpu_duration_ns);
        printf( " s.\n");
    }

#else
    if( quietMode == 0 )
        fprintf(stdout, "\n Encoding process CPU load Time: %4.4f sec ", (float)total_cpu_duration / CLOCKS_PER_SEC);
#endif



    /*------------------------------------------------------------------------------------------*
     * Close files and free ressources
     *------------------------------------------------------------------------------------------*/
    E_IF_evs_close(hEncState);
    // make sure to call close to free malloc'ed memory
    if ( hEncState ) free( hEncState );
    if ( hEncScratch) free( hEncScratch);
    if ( f_input ) fclose( f_input );
    if ( f_stream ) fclose( f_stream );

    if( quietMode == 0 )
        fprintf(stdout, "\n...Done... \n");


    return 0;
}


