/*--------------------------------------------------------------------------------*
 *                                                                                *
 * CodecPro Incorporated                                                          *
 * 905-2160 Terry-Fox                                                             *
 * Laval (Quebec)                                                                 *
 * Canada, H7T 3B7                                                                *
 *                                                                                *
 * support@codecpro.com                                                           *
 * www.codecpro.com                                                               *
 *                                                                                *
 * This material is subject to continuous developments and improvements. All      *
 * warranties implied or expressed, including but not limited to, implied         *
 * warranties of merchantability, fitness for purpose, condition of title,        *
 * noninfringement, are excluded. In no event shall CodecPro Incorporated and its *
 * suppliers be liable for any special, indirect or consequential damages or any  *
 * damages whatsoever arising out of or in connection with the use of this        *
 * information. The foregoing disclaimer shall apply to the maximum extent        *
 * permitted by applicable law, even if any remedy fails its essential purpose.   *
 *--------------------------------------------------------------------------------*/

/*=================================================================================
    EVS Codec 3GPP TS26.442 (see disclaimer.h for version number)
 =================================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "evs_if.h"
#include "io_dec.h"
#include "wavefiletools.h"

/*---------------------------------------------------------------------*
* Local functions
*---------------------------------------------------------------------*/

static void usage_dec(void);
static char *to_upper( char *str );

/*---------------------------------------------------------------------*
 * io_ini_dec()
 *
 * Processing of command line parameters
 *---------------------------------------------------------------------*/

Word32 io_ini_dec(
    const Word32 argc,             /* i  : command line arguments number             */
    char *argv[],                  /* i  : command line arguments                    */
    FILE   **f_stream,             /* o  : input bitstream file                      */
    FILE   **f_synth,              /* o  : output synthesis file                     */
    Word16 *quietMode,             /* o  : limited printouts                         */
    Word16 *noDelayCmp,            /* o  : turn off delay compensation               */
    Word16 *Opt_VOIP,
    Word32 *output_Fs,
    Word16 *bitstreamformat,
    Word16 *amrwb_rfc4867_flag,
    Word16 *sdp_hf_only
)
{
    Word16 i;
    Word16 evs_magic, amrwb_magic;
    Word8 buf[13];



    /*-----------------------------------------------------------------*
     * Initialization
     *-----------------------------------------------------------------*/

    i = 1;
    *f_synth = NULL;
    *f_stream = NULL;
    *Opt_VOIP = 0;

    *bitstreamformat = MIME;
    *amrwb_rfc4867_flag = -1;

    if ( argc <= 1 )
    {
        usage_dec();
        return (-1);
    }

    /*-----------------------------------------------------------------*
     * Optional input arguments
     *-----------------------------------------------------------------*/

    while ( i < argc-3 )
    {
        /*-----------------------------------------------------------------*
         * VOIP mode
         *-----------------------------------------------------------------*/

        if ( strcmp( to_upper(argv[i]), "-VOIP") == 0)
        {
            *Opt_VOIP = 1;
            i += 1;
        }
        else if ( strcmp( to_upper(argv[i]), "-VOIP_HF_ONLY=0") == 0)
        {
            *Opt_VOIP = 1;
            *bitstreamformat = VOIP_RTPDUMP;
            *sdp_hf_only = 0;
            i += 1;
        }
        else if ( strcmp( to_upper(argv[i]), "-VOIP_HF_ONLY=1") == 0)
        {
            *Opt_VOIP = 1;
            *bitstreamformat = VOIP_RTPDUMP;
            *sdp_hf_only = 1;
            i += 1;
        }

        /*-----------------------------------------------------------------*
         * VOIP tracefile
         *-----------------------------------------------------------------*/

        else if ( strcmp( to_upper(argv[i]), "-TRACEFILE" ) == 0 )
        {
            i = i + 2;
        }

        /*-----------------------------------------------------------------*
        * FEC offset file
        *-----------------------------------------------------------------*/

        else if ( strcmp( to_upper(argv[i]), "-FEC_CFG_FILE" ) == 0 )
        {
            i = i + 2;
        }

        /*-----------------------------------------------------------------*
         * Quiet mode
         *-----------------------------------------------------------------*/

        else if ( strcmp( to_upper(argv[i]), "-Q" ) == 0 )
        {
            *quietMode = 1;
            i++;
        }

        /*-----------------------------------------------------------------*
         * deactivate delay compensation
         *-----------------------------------------------------------------*/

        else if ( strcmp( to_upper(argv[i]), "-NO_DELAY_CMP" ) == 0 )
        {
            *noDelayCmp = 1;
            i++;
        }


        /*-----------------------------------------------------------------*
         * Option not recognized
         *-----------------------------------------------------------------*/

        else
        {
            fprintf(stderr, "Error: Unknown option %s\n\n", argv[i]);
            usage_dec();
            return (-1);        }

    } /* end of while  */

    /*-----------------------------------------------------------------*
     * Mandatory input arguments
     *-----------------------------------------------------------------*/

    if (*quietMode == 0)
    {
       printf("==========================================================================\n");
       printf("\n CodecPro EVS v.%d.%d.%d\n", evs_major_version, evs_minor_version, evs_build_version);
       printf("==========================================================================\n\n\n");
    }

    /*-----------------------------------------------------------------*
     * Output sampling frequency
     *-----------------------------------------------------------------*/

    if( i < argc - 2 )
    {
        *output_Fs = (Word32)atoi( argv[i] ) * 1000;
        if( *output_Fs != 8000 && *output_Fs != 16000 && *output_Fs != 32000 && *output_Fs != 48000 )
        {
            fprintf(stderr, "Error: %d kHz is not a supported sampling rate\n\n", atoi( argv[i] ) );
            usage_dec();
            return (-1);
        }

        i++;
    }
    else
    {
        fprintf (stderr, "Error: Sampling rate is not specified\n\n");
        usage_dec();
        return (-1);
    }
    /*-----------------------------------------------------------------*
     * Input bitstream file
     *-----------------------------------------------------------------*/

    if( i < argc - 1 )
    {
        if ( (*f_stream = fopen(argv[i], "rb")) == NULL)
        {
            fprintf(stderr,"Error: input bitstream file %s cannot be opened\n\n", argv[i]);
            usage_dec();
            return (-1);
        }
        /* If MIME/storage format selected, scan for the magic number at the beginning of the bitstream file */
        if((*Opt_VOIP == 0) && (*bitstreamformat == MIME) )
        {
            evs_magic   = 1 ;
            amrwb_magic = 1;

            if(NULL == fgets((char *)buf, 13, *f_stream))
            {
                fprintf(stderr,"Error: input bitstream file %s cannot be read\n\n", argv[i]);
                usage_dec();
                return (-1);
            }

            /* verify AMRWB magic number */
            if ( strncmp((char *)buf, AMRWB_MAGIC_NUMBER, strlen(AMRWB_MAGIC_NUMBER)))
            {
                amrwb_magic = 0;
            }

            if ( strncmp((char *)buf, EVS_MAGIC_NUMBER, strlen(EVS_MAGIC_NUMBER))) /* strncmp safer than strcmp */
            {
                evs_magic = 0;
            }

            if( evs_magic != 0 )
            {
                if ((fread(&buf,sizeof(char), 4, *f_stream) != 4 ) || !((buf[3] == 1) && (buf[2] == 0) && (buf[1] == 0) &&  (buf[0] == 0)) )
                {
                    fprintf(stderr, "Error: input bitstream file %s specifies unsupported number of evs audio channels\n\n",argv[i]);
                    usage_dec();
                    return (-1);
                }
            }

            if( evs_magic == 0 &&  amrwb_magic == 0 )
            {
                /* no valid MIME magic number  */
                fprintf(stderr, "Error: input bitstream file %s specifies unsupported MIME magic number (%13s) \n\n",argv[i],buf );
                usage_dec();
                return (-1);
            }

            if( evs_magic )
            {
                if (*quietMode == 0)
                    fprintf( stdout, "Found MIME Magic number %s\n", EVS_MAGIC_NUMBER );
                *amrwb_rfc4867_flag = 0;
            }
            else
            {
                if (*quietMode == 0)
                    fprintf( stdout, "Found MIME Magic number %s\n",AMRWB_MAGIC_NUMBER );
                *amrwb_rfc4867_flag = 1;
            }

            if (*quietMode == 0)
                fprintf( stdout, "Input bitstream file:   %s\n", argv[i]);

        }

        i++;

    }
    else
    {
        fprintf (stderr, "Error: no input bitstream file specified\n\n");
        usage_dec();
        return (-1);
    }

    /*-----------------------------------------------------------------*
     * Output synthesis file
     *-----------------------------------------------------------------*/

    if( i < argc )
    {
        size_t str_l = strlen(argv[i]);
        if ((str_l >= 5) &&
            (strcmp(to_upper(argv[i] + str_l - 4), ".WAV") == 0))
        {
            Word16 numberOfChannels = 1;
            Word16 bitsPerSample = 16;
            Word32 dataSize; // not used for file writing
            *f_synth = Wave_Fopen(argv[i], "wb+",
                                  &numberOfChannels,
                                  output_Fs,
                                  &bitsPerSample,
                                  &dataSize);
        }

        else *f_synth = fopen(argv[i], "wb");

        if ( *f_synth == NULL )
        {
            fprintf( stderr, "Error: ouput synthesis file %s cannot be opened\n\n", argv[i] );
            usage_dec();
            return (-1);
        }
        if (*quietMode == 0)
            fprintf( stdout, "Output synthesis file:  %s\n", argv[i] );
        i++;
    }
    else
    {
        fprintf( stderr, "Error: no output synthesis file specified\n\n" );
        usage_dec();
        return (-1);
    }
    if (*quietMode == 0)
        fprintf( stdout, "\n" );

    if( !*Opt_VOIP )
    {
        /*-----------------------------------------------------------------*
         * Read information from bitstream
         *-----------------------------------------------------------------*/
        if( *bitstreamformat == MIME )
        {
            if( *amrwb_rfc4867_flag != 0 )
            {
                fseek(*f_stream,(Word32)strlen(AMRWB_MAGIC_NUMBER), SEEK_SET);    /* restart after 9 bytes */
            }
            else
            {
                fseek(*f_stream,(Word32)strlen(EVS_MAGIC_NUMBER)+4, SEEK_SET); /* restart after  16 bytes */
            }
        }
    }

    /*-----------------------------------------------------------------*
        * Print info on screen
        *-----------------------------------------------------------------*/
    /*-----------------------------------------------------------------*
        * Print output sampling frequency
        *-----------------------------------------------------------------*/

    if (*quietMode == 0)
        fprintf( stdout, "Output sampling rate:   %d Hz\n", *output_Fs );

    return (0);
}

/*---------------------------------------------------------------------*
 * to_upper()
 *
 * Capitalize all letters of a string.
 * (normally to_upper() function would be used but it does not work in Unix)
 *---------------------------------------------------------------------*/

static char *to_upper( char *str )
{
    short i;
    char *p = str;

    i = 0;
    while (str[i] != 0)
    {
        if (str[i] >= 'a' && str[i] <= 'z') str[i] -= 0x20;
        i++;
    }

    return p;
}

static void usage_dec( void )
{
    fprintf(stdout,"Usage : EVS_dec.exe [Options] Fs bitstream_file output_file\n\n");

    fprintf(stdout,"Mandatory parameters:\n");
    fprintf(stdout,"---------------------\n");
    fprintf(stdout, "Fs                 : Output sampling rate in kHz (8, 16, 32 or 48)\n");
    fprintf(stdout,"bitstream_file      : Input bitstream filename or RTP packet filename (in VOIP mode)\n");
    fprintf(stdout, "output_file        : Output audio filename (*.8k, *.16k, *.32k, *.48k)\n\n");

    fprintf(stdout,"Options:\n");
    fprintf(stdout,"--------\n");

    fprintf(stdout, "                   : Mime bitstream file format when no VOIP option \n");
    fprintf(stdout, "                     The decoder is able to read both TS26.445 Annex A.2.6 and RFC4867 Mime Storage\n");
    fprintf(stdout, "                     Format files, the magic word in the mime file is used to determine\n");
    fprintf(stdout, "                     which of the two supported formats is in use.\n");
    fprintf(stdout, "                     default bitstream file format is MIME - TS26.445 Annex 2.6\n");
    fprintf(stdout, "-VOIP              : VoIP mode: RTP in MIME\n");
    fprintf(stdout, "-VOIP_hf_only=0    : VoIP mode: EVS RTP Payload Format hf_only=0 in rtpdump\n");
    fprintf(stdout, "-VOIP_hf_only=1    : VoIP mode: EVS RTP Payload Format hf_only=1 in rtpdump\n");
    fprintf(stdout, "                     The decoder may read rtpdump files containing TS26.445 Annex A.2.2\n");
    fprintf(stdout, "                     EVS RTP Payload Format. The SDP parameter hf_only is required.\n");
    fprintf(stdout, "                     Reading RFC4867 AMR/AMR-WB RTP payload format is not supported.\n");
    fprintf(stdout, "-no_delay_cmp      : Turn off delay compensation\n");
    fprintf(stdout, "-q                 : Quiet mode, no frame counter\n");
    fprintf(stdout, "                     default is deactivated\n");
    fprintf(stdout, "\n");
}

