/*--------------------------------------------------------------------------------*
 *                                                                                *
 * CodecPro Incorporated                                                          *
 * 905-2160 Terry-Fox                                                             *
 * Laval (Quebec)                                                                 *
 * Canada, H7T 3B7                                                                *
 *                                                                                *
 * support@codecpro.com                                                           *
 * www.codecpro.com                                                               *
 *                                                                                *
 * This material is subject to continuous developments and improvements. All      *
 * warranties implied or expressed, including but not limited to, implied         *
 * warranties of merchantability, fitness for purpose, condition of title,        *
 * noninfringement, are excluded. In no event shall CodecPro Incorporated and its *
 * suppliers be liable for any special, indirect or consequential damages or any  *
 * damages whatsoever arising out of or in connection with the use of this        *
 * information. The foregoing disclaimer shall apply to the maximum extent        *
 * permitted by applicable law, even if any remedy fails its essential purpose.   *
 *--------------------------------------------------------------------------------*/

/*=================================================================================
      EVS Codec 3GPP TS26.442 (see disclaimer.h for version number)
 =================================================================================*/

#ifndef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "evs_if.h"
#include "io_enc.h"


/* Conversion of bandwidth string into numerical constant */
#define CONV_BWIDTH(bw)                       ( !strcmp(bw, "NB") ? NB : !strcmp(bw, "WB") ? WB : !strcmp(bw, "SWB") ? SWB : !strcmp(bw, "FB") ? FB : -1)


/*-------------------------------------------------------------------*
 * Local functions
 *-------------------------------------------------------------------*/

static void usage_enc(void);
static char *to_upper( char *str );

static char * bit_rate_to_string(char *string, Word32 bit_rate)
{
    char* src, * dst;

    if (!(bit_rate >= 100))
        string = "unsupported bitrate";
    else {
        src = string + sprintf(string, "%i", bit_rate);
        /* Insert a '.' before last two digits and remove last digit */
        /* What we want is to print %.2f of bit_rate/1000.0 */
        dst = src--;
        *--dst = *--src;
        *--dst = *--src;
        *--dst = '.';
    }

    return string;
}

/*---------------------------------------------------------------------*
 * io_ini_enc()
 *
 * Processing of command line parameters
 *---------------------------------------------------------------------*/

Word32 io_ini_enc(
    const Word32   argc,       /* i  : command line arguments number             */
    char  *argv[],             /* i  : command line arguments                    */
    FILE  **f_input,           /* o  : input signal file                         */
    FILE  **f_stream,          /* o  : output bitstream file                     */
    Word16 *quietMode,         /* o  : limit printouts                           */
    Word16 *noDelayCmp,        /* o  : turn off delay compensation               */
    Word32 *input_Fs,
    Word32 *total_brate,       // total bitrate in kbps of the codec
    Word16 *bitstreamformat,   // Bitstream format flag (G192(0), MIME(1) or VOIP_RTPDUMP(2))
    Word16 *Opt_DTX_ON,        // flag indicating DTX operation
    Word16 *Opt_RF_ON,
    Word16 *rf_fec_offset,
    Word16 *rf_fec_indicator,
    Word16 *max_bwidth,        // maximum encoded bandwidth
    Word16 *interval_SID       // CNG and DTX - interval of SID update
)
{
    short i = 1;
    short max_bwidth_user = -1;
    char  max_bwidth_string[4];
    Word32 tmp;
    char stmp[FILENAME_MAX];
    short rf_file_attempt = 1;
    char rf_file_name[300];
    char bit_rate_string[14];

    /*-----------------------------------------------------------------*
     * Initialization
     *-----------------------------------------------------------------*/
    /*The complexity don't need to be taken into account given that is only initialisation */

    *f_input = NULL;
    *f_stream = NULL;
    *quietMode = 0;
    *noDelayCmp = 0;

    *input_Fs = 16000;
    *total_brate = ACELP_12k65;

    *Opt_DTX_ON = 0;
    *Opt_RF_ON = 0;
    *rf_fec_offset = 0;
    *rf_fec_indicator = 1;

    *max_bwidth = SWB;
    *interval_SID = FIXED_SID_RATE;
    *bitstreamformat = MIME;

    if ( argc < 5 )
    {
        usage_enc();
        return (-1);
    }


    /*-----------------------------------------------------------------*
     * Optional input arguments
     *-----------------------------------------------------------------*/

    while ( i < argc-4 ){
        /*-----------------------------------------------------------------*
         * Bandwidth limitation
         *-----------------------------------------------------------------*/

        if ( strcmp( to_upper(argv[i]), "-MAX_BAND") == 0 ){
            strncpy( stmp, argv[i+1], sizeof(stmp) );

            if (strcmp( to_upper(stmp), "-NB") == 0 || strcmp( to_upper(stmp), "NB") == 0){
                *max_bwidth = max_bwidth_user = NB;
            }
            else if (strcmp( to_upper(stmp), "-WB") == 0 || strcmp( to_upper(stmp), "WB") == 0){
                *max_bwidth = max_bwidth_user = WB;
            }
            else if (strcmp( to_upper(stmp), "-SWB") == 0 || strcmp( to_upper(stmp), "SWB") == 0){
                *max_bwidth = max_bwidth_user = SWB;
            }
            else if (strcmp( to_upper(stmp), "-FB") == 0 || strcmp( to_upper(stmp), "FB") == 0){
                *max_bwidth = max_bwidth_user = FB;
            }
            else {
                fprintf(stderr, "Error: incorect bandwidth specification\n\n");
                usage_enc();
                return (-1);
            }
            if (*quietMode == 0) {
                fprintf(stdout, "Maximum encoded bandwidth: %s\n", stmp);
            }

            i += 2;
        }

        /*-----------------------------------------------------------------*
         * Quiet mode
         *-----------------------------------------------------------------*/

        else if( strcmp( to_upper(argv[i]), "-Q" ) == 0 )
        {
            i++;
            *quietMode = 1;
        }

        /*-----------------------------------------------------------------*
         * DTX/CNG
         *-----------------------------------------------------------------*/

        else if( strcmp( to_upper(argv[i]), "-DTX" ) == 0 ) {
            i++;
            if( i < argc-4 ) {
                if( sscanf(argv[i], "%d", &tmp) <= 0 ) {
                    tmp = FIXED_SID_RATE;
                }
                else {
                    i++;
                }
            }
            else {
                tmp = FIXED_SID_RATE;
            }

            *Opt_DTX_ON = 1;
            if( tmp == 0 ) {
                *interval_SID = 0;
            }
            else if (tmp >= 3 && tmp <= 100) {
                *interval_SID = (short)tmp;
            }
            else {
                fprintf(stderr, "Error: Incorrect SID update interval specified: %d (supported 3-100)\n\n", tmp);
                usage_enc();
                return (-1);
            }
        }

        /*-----------------------------------------------------------------*
         * Deactivate delay compensation
         *-----------------------------------------------------------------*/
        else if( strcmp( to_upper(argv[i]), "-NO_DELAY_CMP" ) == 0 ) {
            *noDelayCmp = 1;
            i++;
        }



        /*-----------------------------------------------------------------*
         * Activate channel-aware mode
         *-----------------------------------------------------------------*/

        else if ( strcmp( to_upper(argv[i]), "-RF" ) == 0 ) {
            *Opt_RF_ON = 1;
            i++;
            *rf_fec_indicator = 1;
            *rf_fec_offset = FEC_OFFSET;
        }

        /*-----------------------------------------------------------------*
         * Option not recognized
         *-----------------------------------------------------------------*/

        else
        {
            fprintf(stderr, "Error: option not recognized, %s\n\n", argv[i]);
            usage_enc();
            return (-1);
        }
    } /* end of while  */

    /*-----------------------------------------------------------------*
     * Mandatory input arguments
     *-----------------------------------------------------------------*/

    if( *quietMode == 0 ) {
	    printf("==========================================================================\n");
	    printf("\n CodecPro EVS v.%d.%d.%d\n", evs_major_version, evs_minor_version, evs_build_version);
	    printf("==========================================================================\n\n\n");
    }
	
    /*-----------------------------------------------------------------*
     * Bitrate
     *-----------------------------------------------------------------*/

    if ( i < argc-2 )
    {
        /* check if profile file has been entered instead of a fixed bitrate */
        sscanf(argv[i], "%d", &tmp);
        *total_brate = tmp;

        /* channel-aware mode is supported only at 13.20 */
        if( *Opt_RF_ON && *total_brate != ACELP_13k20 ) {
            *Opt_RF_ON = 0;
        }

        /* SC-VBR at 5.90 kbps */
        if ( *total_brate == ACELP_5k90 ) {

            if ( *max_bwidth != NB ) {
                *max_bwidth = WB;
            }
        }


        /* check if the entered bitrate is supported */
            switch ( *total_brate ) {
                case ACELP_6k60 :
                case ACELP_8k85 :
                case ACELP_12k65 :
                case ACELP_14k25 :
                case ACELP_15k85 :
                case ACELP_18k25 :
                case ACELP_19k85 :
                case ACELP_23k05 :
                case ACELP_23k85 :
                case ACELP_5k90  :
                case ACELP_7k20  :
                case ACELP_8k00  :
                case ACELP_9k60  :
                case ACELP_13k20 :
                case ACELP_16k40 :
                case ACELP_24k40 :
                case ACELP_32k   :
                case ACELP_48k   :
                case ACELP_64k   :
                case HQ_96k      :
                case HQ_128k     :
                    break;
                default : {
                    fprintf(stderr, "Error1: Incorrect bitrate specification: %d\n\n", *total_brate );
                    usage_enc();
                    return (-1);
                }
                break;
            }
            i++;
    }
    else {
        fprintf(stderr, "Error: no bitrate specified\n\n");
        usage_enc();
        return (-1);
    }

    /*-----------------------------------------------------------------*
     * Input sampling frequency
     *-----------------------------------------------------------------*/

    if( i < argc-2 ) {
        *input_Fs = (Word32) atoi( argv[i] ) * 1000;
        if( *Opt_RF_ON && *input_Fs == 8000 ) {
            /* Reset RF parameters if NB input_Fs;
               For FB input @13.2kbps, the codec is shifted down to SWB below. */
            *Opt_RF_ON = 0;
            *rf_fec_offset = 0;
        }

        if( (*input_Fs != 8000) && ( *input_Fs != 16000) && (*input_Fs != 32000) && (*input_Fs != 48000) ) {
            fprintf(stderr, "Error: %d kHz is not a supported sampling rate\n\n", atoi( argv[i] ) );
            usage_enc();
            return (-1);
        }

        i++;
    }
    else {
        fprintf(stderr, "Error: no input sampling frequency specified\n\n");
        usage_enc();
        return (-1);
    }

    /*-----------------------------------------------------------------*
     * Input file
     *-----------------------------------------------------------------*/

    if( i < argc-1 ) {
        if ( (*f_input = fopen(argv[i], "rb")) == NULL ) {
            fprintf(stderr, "Error: input audio file %s could not be opened\n\n", argv[i]);
            usage_enc();
            return (-1);
        }

        if( *quietMode == 0 )
            fprintf(stdout, "Input audio file:       %s\n", argv[i]);
        i++;
    }
    else {
        fprintf(stderr, "Error: no input file specified\n\n");
        usage_enc();
        return (-1);
    }

    /*-----------------------------------------------------------------*
     * Output bitstream file
     *-----------------------------------------------------------------*/

    if( i < argc ) {
        if ( (*f_stream = fopen(argv[i], "wb")) == NULL) {
            fprintf(stderr, "Error: output bitstream file %s could not be opened\n\n", argv[i]);
            usage_enc();
            return (-1);
        }
        if( *quietMode == 0 )
            fprintf(stdout, "Output bitstream file:  %s\n", argv[i]);
        i++;
        /* If MIME/storage format selected, write the magic number at the beginning of the bitstream file */
        if( *bitstreamformat == MIME ) {
#ifdef EVS_MAGIC_NUMBER
            char buf[4];
            fwrite(EVS_MAGIC_NUMBER, sizeof(char), strlen(EVS_MAGIC_NUMBER), *f_stream);
            buf[0] = buf[1] = buf[2] = 0;
            buf[3] = 1;
            fwrite(&buf, sizeof(char), 4, *f_stream);
            if( *quietMode == 0 )
                fprintf(stdout, "Output bitstream file format: MIME");
#endif
        }
    }
    else {
        fprintf(stderr, "Error: no output bitstream file specified\n\n");
        usage_enc();
    }

    if( *quietMode == 0 )
        fprintf( stdout, "\n" );


    /* Prevent st->max_bwidth from being higher than Fs/2 */
    if ( *input_Fs == 8000 && *max_bwidth > NB ) {
        *max_bwidth = NB;
    }
    else if ( *input_Fs == 16000 && *max_bwidth > WB ) {
        *max_bwidth = WB;
    }
    else if ( *input_Fs == 32000 && *max_bwidth > SWB ) {
        *max_bwidth = SWB;
    }

    /*-----------------------------------------------------------------*
     * Print info on screen
     *-----------------------------------------------------------------*/
    /*-----------------------------------------------------------------*
     * Print input signal sampling frequency
     *-----------------------------------------------------------------*/

    if( *quietMode == 0 )
        fprintf(stdout, "Input sampling rate:    %d Hz\n", *input_Fs);

    /*-----------------------------------------------------------------*
     * Print bitrate
     *-----------------------------------------------------------------*/

    if( *quietMode == 0 ) {
        if (*total_brate == ACELP_5k90 ) {
            fprintf(stdout, "Average bitrate:        %s kbps\n", bit_rate_to_string(bit_rate_string, ACELP_5k90));
        }
        else {
            fprintf(stdout, "Bitrate:                %s kbps\n", bit_rate_to_string(bit_rate_string, (Word32)*total_brate));
        }
    }

    /*-----------------------------------------------------------------*
     * Print CNG update interval, if DTX is activated
     *-----------------------------------------------------------------*/

    if( *quietMode == 0 ) {
        if ( *Opt_DTX_ON )
        {
            if( !*interval_SID ) {
                fprintf(stdout, "DTX:                    ON, variable CNG update interval\n");
            }
            else {
                fprintf(stdout, "DTX:                    ON, CNG update interval = %d frames\n", *interval_SID);
            }
        }
    }

    /*-----------------------------------------------------------------*
     * Print channel-aware mode info
     *-----------------------------------------------------------------*/

    if( *quietMode == 0 ) {
        if ( *Opt_RF_ON )
        {
            fprintf(stdout, "Channel-aware mode:     ON, Channel-aware config file name:  %s  \n", rf_file_name );
        }
    }

    /*-----------------------------------------------------------------*
     * Print potential limitation of audio bandwidth
     *-----------------------------------------------------------------*/
    switch(*max_bwidth)
    {
    case NB:
        strncpy(max_bwidth_string, "NB\0", sizeof(max_bwidth_string));
        break;
    case WB:
        strncpy(max_bwidth_string, "WB\0", sizeof(max_bwidth_string));
        break;
    case SWB:
        strncpy(max_bwidth_string, "SWB\0", sizeof(max_bwidth_string));
        break;
    case FB:
        strncpy(max_bwidth_string, "FB\0", sizeof(max_bwidth_string));
        break;
    default:
        memset(max_bwidth_string, 0, sizeof(max_bwidth_string));
        break;
    }

    if ((*total_brate == ACELP_5k90 ) && !*Opt_DTX_ON)
    {
        fprintf(stderr, "\nError: SC-VBR 5900 bit/s not supported without DTX\n\n");
        return(-1);
    }

    if (*quietMode == 0)
    {
        if ((max_bwidth_user != -1) && (*max_bwidth != max_bwidth_user)) {
            fprintf(stdout, "\nBandwidth limited to %s.\n", max_bwidth_string);
        }
        if ((max_bwidth_user == -1) && (*max_bwidth < FB) && (*input_Fs == 48000)) {
            fprintf(stdout, "\nBandwidth limited to %s. To enable FB coding, please use -max_band FB.\n", max_bwidth_string);
        }
        if ((*max_bwidth == FB) && (*total_brate < ACELP_16k40)) {
            fprintf(stdout, "\nFB coding not supported below %s kbps. ", bit_rate_to_string(bit_rate_string, ACELP_16k40));
            if (*total_brate < ACELP_9k60) {
                fprintf(stdout, "Switching to WB.\n");
            }
            else {
                fprintf(stdout, "Switching to SWB.\n");
            }
        }
        if ((*max_bwidth == SWB) && (*total_brate < ACELP_9k60)) {
            fprintf(stdout, "\nSWB coding not supported below %s kbps. Switching to WB.", bit_rate_to_string(bit_rate_string, ACELP_9k60));
        }
        fprintf(stdout, "\n");
    }

    /* in case of 8kHz input sampling or "-max_band NB", require the total bitrate to be below 24.40 kbps */
    if ( ((*max_bwidth == NB) || (*input_Fs == 8000)) && (*total_brate > ACELP_24k40) ) {
        fprintf(stderr, "\nError: Unsupported mode NB %d bit/s, NB mode supports rates 5900-24400 bit/s\n\n", *total_brate);
        usage_enc();
    }

    return (0);
}

/*---------------------------------------------------------------------*
 * read_next_rfparam()
 *
 * Read next channel aware configuration parameters
 *      p: RF FEC indicator HI or LO, and
 *      o: RF FEC offset in number of frame (2, 3, 5, or 7)
 *---------------------------------------------------------------------*/

Word32 read_next_rfparam(
    Word16 *rf_fec_offset,    /* o: rf offset                         */
    Word16 *rf_fec_indicator, /* o: rf FEC indicator                  */
    FILE* f_rf                /* i: file pointer to read parameters   */
)
{
    char rline[10], str[4];
    Word32 tmp;

    /* initialize */
    *rf_fec_offset = 0;
    *rf_fec_indicator = 1;


    if( f_rf != NULL )
    {

        while ( fgets(rline, 10, f_rf) == NULL && feof(f_rf) )
        {
            rewind(f_rf);
        }
    }
    else
    {
        return(0);
    }

    if ( sscanf (rline,"%s %d",str,&tmp) != 2)
    {
        fprintf(stderr, "Error in the RF configuration file. There is no proper configuration line.\n");
        return(-1);
    }

    /* Read RF FEC indicator */
    if ( strcmp( to_upper(str), "HI" ) == 0 )
    {
        *rf_fec_indicator = 1;
    }
    else if ( strcmp( to_upper(str), "LO" ) == 0 )
    {
        *rf_fec_indicator = 0;
    }
    else
    {
        fprintf(stderr, " Incorrect FEC indicator string. Exiting the encoder.\n");
        return(-1);
    }

    /* Read RF FEC offset */
    if( tmp == 0 || tmp == 2 || tmp == 3 || tmp == 5 || tmp == 7 )
    {
        *rf_fec_offset = (Word16)tmp;
    }
    else
    {
        fprintf (stderr,"Error: incorrect FEC offset specified in the RF configration file; RF offset can be 2, 3, 5, or 7. \n");
        return(-1);
    }
    return (0);
}




/*---------------------------------------------------------------------*
 * read_next_brate()
 *
 * Read next bitrate from profile file (only if invoked on cmd line)
 *---------------------------------------------------------------------*/

Word32 read_next_brate(
    Word32  *total_brate,             /* i/o: total bitrate                             */
    FILE   *f_rate                   /* i  : bitrate switching profile (0 if N/A)      */
)
{
    UWord8 total_brate_array[4];

    /*The complexity don't need to be taken into account given that is only simulation */
    /* read next bitrate value from the profile file */
    if( f_rate != NULL ) {
        while ( fread(total_brate_array, 4, 1, f_rate ) != 1 && feof(f_rate) ) {
            rewind(f_rate);
        }
			
        // modified to accomodate bid endian
        *total_brate = total_brate_array[0] + ((total_brate_array[1] << 8) & 0xff00) +
                     ((total_brate_array[2] << 16) & 0xff0000) + ((total_brate_array[3] << 24) & 0xff000000);
    }

    return(0);
}

/*---------------------------------------------------------------------*
 * read_next_bwidth()
 *
 * Read next bandwidth from profile file (only if invoked on cmd line)
 *---------------------------------------------------------------------*/

Word32 read_next_bwidth(
    Word16  *max_bwidth,            /* i/o: maximum encoded bandwidth                 */
    FILE    *f_bwidth,              /* i  : bandwidth switching profile (0 if N/A)    */
    Word32  *bwidth_profile_cnt,    /* i/o: counter of frames for bandwidth switching profile file */
    Word32   input_Fs               /* i  : input sampling frequency                  */
)
{
    Word32 res;
    char stmp[4];

    if ( *bwidth_profile_cnt == 0 )
    {
        /* read next bandwidth value and number of frames from the profile file */
        while ( (res = fscanf( f_bwidth, "%d %3s", bwidth_profile_cnt, stmp)) != 2 && feof(f_bwidth))
        {
            rewind(f_bwidth);
        }

        (*bwidth_profile_cnt)--;

        if (strcmp( to_upper(stmp), "NB") == 0 )
        {
            *max_bwidth = NB;
        }
        else if (strcmp( to_upper(stmp), "WB") == 0 )
        {
            *max_bwidth = WB;
        }
        else if (strcmp( to_upper(stmp), "SWB") == 0 )
        {
            *max_bwidth = SWB;
        }
        else if (strcmp( to_upper(stmp), "FB") == 0 )
        {
            *max_bwidth = FB;
        }

        /* Prevent max_bwidth from being higher than Fs/2 */
        if ( input_Fs == 8000 && *max_bwidth > NB )
        {
            *max_bwidth = NB;
        }
        else if ( input_Fs == 16000 && *max_bwidth > WB )
        {
            *max_bwidth = WB;
        }
        else if ( input_Fs == 32000 && *max_bwidth > SWB )
        {
            *max_bwidth = SWB;
        }

    }
    else
    {
        /* current profile still active, only decrease the counter */
        (*bwidth_profile_cnt)--;
    }

    return(0);
}



/*---------------------------------------------------------------------*
 * to_upper()
 *
 * Capitalize all letters of a string.
 * (normally, _strupr() function would be used but it does not work in Unix)
 *---------------------------------------------------------------------*/

static char *to_upper( char *str )
{
    short i;

    i = 0;
    while (str[i] != 0)
    {
        if (str[i] >= 'a' && str[i] <= 'z') str[i] += 'A' - 'a';
        i++;
    }

    return str;
}

static void usage_enc( void )
{
    fprintf(stdout, "Usage: EVS_cod.exe [Options] R Fs input_file bitstream_file\n\n");

    fprintf(stdout, "Mandatory parameters:\n");
    fprintf(stdout, "---------------------\n");
    fprintf(stdout, "R                : Bitrate in bps, \n");
    fprintf(stdout, "                   for EVS native modes R = (5900*, 7200, 8000, 9600, 13200, 16400,\n");
    fprintf(stdout, "                                             24400, 32000, 48000, 64000, 96000, 128000) \n");
    fprintf(stdout, "                                             *VBR mode (average bitrate),\n");
    fprintf(stdout, "                   for AMR-WB IO modes R =  (6600, 8850, 12650, 14250, 15850, 18250,\n");
    fprintf(stdout, "                                             19850, 23050, 23850) \n");
    fprintf(stdout, "                   Alternatively, R can be a bitrate switching file which consits of R values\n");
    fprintf(stdout, "                   indicating the bitrate for each frame in bit/s. These values are stored in\n");
    fprintf(stdout, "                   binary format using 4 bytes per value\n");
    fprintf(stdout, "Fs               : Input sampling rate in kHz, Fs = (8, 16, 32 or 48) \n");

    fprintf(stdout, "input_file       : Input signal filename \n");
    fprintf(stdout, "bitstream_file   : Output bitstream filename \n\n");

    fprintf(stdout, "Options:\n");
    fprintf(stdout, "--------\n");
    fprintf(stdout, "-q               : Quiet mode, no frame counters\n");
    fprintf(stdout, "                 : default is deactivated\n");
    fprintf(stdout, "-dtx D           : Activate DTX mode, D = (0, 3-100) is the SID update rate\n");
    fprintf(stdout, "                   where 0 = adaptive, 3-100 = fixed in number of frames,\n");
    fprintf(stdout, "                   default is deactivated\n");
    fprintf(stdout, "-dtx             : Activate DTX mode with a SID update rate of 8 frames\n");
    fprintf(stdout, "                   default is deactivated\n");
    fprintf(stdout, "-rf p o          : Activate channel-aware mode for WB and SWB signal at 13.2kbps, \n");
    fprintf(stdout, "                   where FEC indicator, p: LO or HI, and FEC offset, o: 2, 3, 5, or 7 in number of frames.\n");
    fprintf(stdout, "                   Alternatively p and o can be replaced by a rf configuration file with each line  \n");
    fprintf(stdout, "                   contains the values of p and o separated by a space, \n");
    fprintf(stdout, "                   default is deactivated \n");
    fprintf(stdout, "-max_band B      : Activate bandwidth limitation, B = (NB, WB, SWB or FB)\n");
    fprintf(stdout, "                   alternatively, B can be a text file where each line contains\n");
    fprintf(stdout, "                   \"nb_frames B\"\n");
    fprintf(stdout, "-no_delay_cmp    : Turn off delay compensation\n");
    #if 0
    fprintf(stdout, "-mime            : Mime output bitstream file format\n");
    fprintf(stdout, "                   The encoder produces TS26.445 Annex.2.6 Mime Storage Format, (not RFC4867 Format)\n");
    fprintf(stdout, "                   default output bitstream file format is MIME\n");
    #endif

    fprintf(stdout, "\n");
}
